print("import: 'pyarrow'")
import pyarrow

print("import: 'pyarrow.compute'")
import pyarrow.compute

print("import: 'pyarrow.csv'")
import pyarrow.csv

print("import: 'pyarrow.dataset'")
import pyarrow.dataset

print("import: 'pyarrow.filesystem'")
import pyarrow.filesystem

print("import: 'pyarrow.flight'")
import pyarrow.flight

print("import: 'pyarrow.json'")
import pyarrow.json

print("import: 'pyarrow.orc'")
import pyarrow.orc

print("import: 'pyarrow.parquet'")
import pyarrow.parquet

print("import: 'pyarrow.plasma'")
import pyarrow.plasma

print("import: 'pyarrow.fs'")
import pyarrow.fs

print("import: 'pyarrow._s3fs'")
import pyarrow._s3fs

print("import: 'pyarrow._hdfs'")
import pyarrow._hdfs

